! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!                      FD4 load balancing                      !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! FD4 load balancing & creation/removal of blocks.
!!
!! Preprocessor options:
!!  - $FD4_METIS: add ability to calculate new partition with ParMETIS.
!!  - $FD4_PARTITION_STATISTICS:
!!    - collect the size of block migration messages *received* in $comm_total_size (module $fd4_comm_mod)
!!    - collect timing of fd4_part_new or fd4_part_metis in $balance_timing_us
!!  - $FD4_INSTANT_BGB: see $fd4_domain
!!
!! Load balancing methods:
!!  - $FD4_BALANCE_HSFC_SIMPLE:   simple Hilbert SFC method, not recommended
!!  - $FD4_BALANCE_HSFC_BINSRCH:  high quality Hilbert SFC method
!!  - $FD4_BALANCE_HSFC_PARALLEL: high quality Hilbert SFC method, parallelized
!!  - $FD4_BALANCE_METIS:         ParMETIS_V3_AdaptiveRepart (only if FD4 is compiled with ParMETIS)
!!  - $FD4_BALANCE_SERIAL:        move all blocks to rank 0 (useful for debugging only)
!!
!! *TODO*:
!!  - move type $fd4_domain_balance and its initialization from $fd4_domain to here ?
!!    - this has impact on user interface and $fd4_util
!!
!! @see fd4_domain
!! @see fd4_part_metis
!! @see fd4_part_sfc
!! @author Matthias Lieber
module fd4_balance_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_block_mod
  use fd4_domain_mod
  use fd4_ghostlist_mod
  use fd4_iter_mod
  use fd4_mpi_mod
  use fd4_part_sfc_mod
#ifdef FD4_METIS
  use fd4_part_metis_mod
#endif
  use fd4_comm_mod
  use fd4_mpitype_mod
  use rbtree_fd4_block_mod
  use stack_mod
  use stack3_mod
  use mrgrnk_mod
  use merge_mod
  use section_mod
  use timing_mod

  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH

  integer(i_k), parameter, public :: FD4_BALANCE_HSFC_SIMPLE = FD4_PART_SFC_SIMPLE
  integer(i_k), parameter, public :: FD4_BALANCE_HSFC_BINSRCH = FD4_PART_SFC_BINSRCH
  integer(i_k), parameter, public :: FD4_BALANCE_HSFC_PARALLEL = FD4_PART_SFC_PARALLEL
  integer(i_k), parameter, public :: FD4_BALANCE_HILBERT = FD4_BALANCE_HSFC_BINSRCH
  integer(i_k), parameter, public :: FD4_BALANCE_METIS = 101
  integer(i_k), parameter, public :: FD4_BALANCE_SERIAL = 102
  
#ifdef FD4_PARTITION_STATISTICS  
  integer(i8k), public :: balance_timing_us
  integer(i8k), public :: check_timing_us
  integer(i8k), public :: migration_timing_us
#endif


  !! type for fd4_balance_readjust to return some statistics information about load balancing
  !! (to be extended)
  type fd4_balance_statistics
    logical      :: partition_changed = .false.   !! global: indicates if partition has been *tried* to be changed
    integer(i_k) :: migrated_blocks = 0           !! global: number of migrated blocks
    integer(i_k) :: removed_blocks = 0            !! global: number of removed blocks
    integer(i_k) :: created_blocks = 0            !! global: number of created blocks
    real(r4k)    :: last_measured_balance = -1.0  !! global: measured load balance based on given block weights
    integer(i_k) :: allocated_blocks = 0          !! local: number of block allocations (incl. ghosts, excl. pool cached)
    integer(i_k) :: deallocated_blocks = 0        !! local: number of block deallocations (incl. ghosts, excl. pool cached)
    character(40):: status                        !! global: tells what $fd4_balance_readjust did and why
    real(r4k)    :: auto_tlost                    !! global: time lost due to imbalance (auto mode only)
    real(r4k)    :: auto_tbalance                 !! global: estimate time spend for load balancing (auto mode only)
  end type fd4_balance_statistics

  public :: fd4_balance_readjust, fd4_balance_statistics
  
contains

  include 'framework/fd4_domain.incf'
  include 'util/stack.incf'
  include 'util/stack3.incf'

  !! Readjust global work distrubution after an iteration over blocks.
  !!
  !! - 1. check if repartitioning is required
  !!  - update global empty info (using fd4_balance_global_empty_update) -> $empty_changed
  !!  - create array of blocks (block ids) which are needed (using $domain%$empties to get global view) -> $blocks_changed
  !!  - check load balance (using $block%weight) -> $balance_changed
  !!  - $part_changed = $balance_changed .or. $blocks_changed
  !! - if ($part_changed)
  !!  - 2. calculate new partioning: 
  !!    - call fd4_part_sfc or fd4_part_metis to compute new local partitioning
  !!  - 3. block migration
  !!    - using fd4_balance_block_migration
  !! - if($part_changed .or. $opt_force)
  !!  - 4. communication ghost block update
  !!    - using fd4_balance_update_ghosts
  !!
  !!
  !! *More info about the optional paramters*:
  !!  - $opt_ech: If true, $fd4_balance_readjust assumes that the empty info did not change in any block. That means,
  !!    no new block needs to be created (or removed). In this case, some global communication can be safed to speed up
  !!    $fd4_balance_readjust. Default is true, unless all variable have no threshold value ($vthres).
  !!    In this case,  $opt_ech will be ignored.
  !!  - $opt_auto: If true, $fd4_balance_readjust decides automatically whether load balancing is beneficial or not.
  !!    This requires that the block weights are set to the total *computation* time of the block in microseconds since
  !!    the last call to $fd4_balance_readjust. $opt_lbtol is ignored if $opt_auto is set to true.
  !!  - $opt_ubvec, $opt_itr, $opt_subset, $opt_wfactor: Only for ParMetis load balancing. $opt_subset is experimental.
  !!  - $opt_weight: default is .true., i.e. block weights are considered. If $opt_weight is set to .false., the auto
  !!    mode does not work.
  !!
  !!
  !! *TODO*:
  !!  - Load imbalance is not measured when blocks change ($opt_ech is false and no block created/removed).
  !!    That means $fd4_balance_statistics%$last_measured_balance is not set.
  !!  - an optional argument to this routine should indicate if the weights have changed,
  !!    if not, we could skip the calls to fd4_balance_estimate
  !!  - *ALWAYS CHECK*: single block case, rank-has-no-block case, single process case, ng == 0 case, no block at all case working?
  !!
  subroutine fd4_balance_readjust(domain, err, method, opt_ubvec, opt_itr,    opt_subset, opt_wfactor, &
                                                       opt_force, opt_lbtol,  opt_stats,  opt_ech,     &
                                                       opt_auto,  opt_weight )
    type(fd4_domain), intent(inout), target :: domain     !! the domain
    integer(i_k),     intent(out)           :: err        !! error status: 0...ok
    integer(i_k),     intent(in), optional  :: method     !! load balancing method to use (see module doc above)
    real(r4k),        intent(in), optional  :: opt_ubvec  !! optional ParMetis ubvec parameter
    real(r4k),        intent(in), optional  :: opt_itr    !! optional ParMetis itr parameter
    integer(i_k),     intent(in), optional  :: opt_subset !! use only $subset processes for ParMetis calculation (<2...off)
    real(r4k),        intent(in), optional  :: opt_wfactor!! ParMetis: multiply this with block's weight before converting to int
    logical,          intent(in), optional  :: opt_force  !! force to run ghost block allocation and deallocation    
    real(r4k),        intent(in), optional  :: opt_lbtol  !! load balance tolerance, 0.0 (max. relaxed) ... 1.0 (strictest)
    type(fd4_balance_statistics), &
                      intent(out),optional  :: opt_stats  !! return some statistics about load balancing
    logical,          intent(in), optional  :: opt_ech    !! if false, assume that empty info did not change in any block
    logical,          intent(in), optional  :: opt_auto   !! automatic decision whether load balancing beneficial or not
    logical,          intent(in), optional  :: opt_weight !! if false ignore block weights (i.e. assume each block has same weight)
    
    type(fd4_mpi), pointer :: mpi
    type(stack) :: mystack                         ! tmp stack
    type(stack3), pointer :: allblocks             ! coordinates of all required blocks in global domain
    integer(i_k), allocatable :: myblocks(:,:)     ! coordinates of local blocks (result from partitioning algorithm)
    logical :: empty_changed, balance_changed, blocks_changed, part_changed, force, auto, lb_benefit, use_weights
    integer(i_k) :: meth, newpartsize
    integer(i8k) :: lbtime0, lbtime1
#ifdef FD4_PARTITION_STATISTICS  
    integer(i8k) :: timing_us0, timing_us1
#endif
#ifdef FD4_METIS
    real(r4k) :: ubvec(1), itr, wfactor
    integer(i_k) :: subset
#endif
    real(r4k) :: balance
    real(r4k) :: balance_tol
#ifdef FD4_BALANCE_DUMP_WEIGHTS
    integer, save :: dumpcounter = 0
    integer(i_k) :: i,j,k
#endif

    err = -1
    mpi => domain%mpi
    empty_changed = .false.
    balance_changed = .false.
    blocks_changed = .false.
    part_changed = .false.
    allblocks => domain%balance%allblocks

#ifdef FD4_PARTITION_STATISTICS
    comm_total_size = 0
    balance_timing_us = 0
    check_timing_us = 0
    migration_timing_us = 0
#endif

    meth = FD4_BALANCE_HSFC_BINSRCH
    if(present(method)) then
      meth = method
#ifdef FD4_METIS
      ! do not run ParMetis in single process case
      if(mpi%nprocs<2) meth = FD4_BALANCE_HSFC_SIMPLE
#else
      FD4_L1_ASSERT(meth==FD4_BALANCE_METIS,'FD4 is not compiled with ParMetis')
#endif
#if ( FD4_VERBOSE_LEVEL >= 1 )
      if(meth/=FD4_BALANCE_METIS  .and. meth/=FD4_BALANCE_HSFC_SIMPLE  .and. &
         meth/=FD4_BALANCE_SERIAL .and. meth/=FD4_BALANCE_HSFC_BINSRCH .and. &
         meth/=FD4_BALANCE_HSFC_PARALLEL) then
        FD4_L1_ERROR('balance method not valid: '//i2s(meth))
      end if
#endif
    end if
    
    force = .false.
    if(present(opt_force)) force = opt_force
    
    balance_tol = 0.95
    if(present(opt_lbtol)) balance_tol = opt_lbtol
    balance = 0.0
    
    use_weights = .true.
    if(present(opt_weight)) use_weights = opt_weight

    empty_changed = .true.
    if(present(opt_ech)) empty_changed = opt_ech 
    empty_changed = empty_changed .and. domain%varinfo%vthres_mode

    auto = .false.
    if(present(opt_auto)) then
      auto = opt_auto
      ! override default or user-given balance_tol
      if(auto) balance_tol = 1.0      
    end if
    lb_benefit = .true.
    
    FD4_L1_ASSERT(auto.and..not.use_weights,'need block weights in auto mode')


#ifdef FD4_PARTITION_STATISTICS  
      call gettime(timing_us0)
#endif

    if(present(opt_stats)) then
      opt_stats%partition_changed     = .false.
      opt_stats%migrated_blocks       = 0
      opt_stats%removed_blocks        = 0
      opt_stats%created_blocks        = 0
      opt_stats%last_measured_balance = -1.0
      opt_stats%allocated_blocks      = 0
      opt_stats%deallocated_blocks    = 0
      opt_stats%status                = ' '
      opt_stats%auto_tlost            = 0.0
      opt_stats%auto_tbalance         = 0.0
    end if

!!
!! 1. check if repartitioning is required
!!

    !! check if empty-info has changed
    if(empty_changed) then
      
      ! caller assumes that empty info may have changed, let's check this
      call fd4_balance_global_empty_update(domain, mystack, empty_changed, lbtime0)

    else
    
      ! force to get the first time after global sync later
      lbtime0 = -1
      
    end if

    !! check if balance is not sufficient
    if(use_weights .and. .not. empty_changed) then

      !! ok, the empty info did not change, but maybe the load
      !! of the processes is not balanced, let's check this
      call fd4_balance_estimate(domain, auto, lb_benefit, balance, lbtime1, opt_stats = opt_stats)
      
      ! if this was the first global sync, we take this time as starting time
      if(lbtime0 < 0) lbtime0 = lbtime1
      
      if(present(opt_stats)) opt_stats%last_measured_balance = balance

      ! in case the balance is poor, we sign the balance as changed to
      ! force rebalance
      if(balance < balance_tol .and. lb_benefit) balance_changed = .true.

    end if    

    ! force full load balancing in case of FD4_BALANCE_SERIAL
    if(meth == FD4_BALANCE_SERIAL) balance_changed = .true.
    

    !! let's see if the number of blocks has changed
    !! and create the list of required blocks
    if(empty_changed .or. domain%blocks_are_marked .or. allblocks%pos==0) then

      !! create array of all required blocks - domain%balance%allblocks
      call fd4_balance_required_blocks(domain, blocks_changed )
      
    else
      
      ! since empty did not change and no blocks were marked:
      blocks_changed = .false.
      ! we can now reuse array allblocks (from last call) for the next steps
      
    end if


    if(use_weights .and. empty_changed .and. .not. blocks_changed) then
    
      !! ok, the number of blocks did not change, but maybe the load
      !! of the processes is not balanced, let's check this
      call fd4_balance_estimate(domain, auto, lb_benefit, balance, opt_stats = opt_stats)
      
      if(present(opt_stats)) opt_stats%last_measured_balance = balance
      
      ! in case the balance is poor, we sign the balance as changed to
      ! force rebalance
      if(balance < balance_tol .and. lb_benefit) balance_changed = .true.
      
    end if    


#ifdef FD4_PARTITION_STATISTICS  
      call gettime(timing_us1)
      check_timing_us = timing_us1 - timing_us0
#endif
    
    ! check if the partitioning need to be changed
    part_changed = blocks_changed .or. balance_changed
    
    ! ...if yes, we go to step 2: calculate the new partitioning
    if(part_changed) then


!!
!! 2. calculate new partioning
!!
      
#if ( FD4_VERBOSE_LEVEL >= 4 )
      if(mpi%rank==0 .and. allblocks%pos==0) then
         FD4_L4_INFO('Total block count = '//i2s(allblocks%pos))
      end if
#endif

      allocate(myblocks(3,allblocks%pos))
      newpartsize = 0

#ifdef FD4_PARTITION_STATISTICS  
      call gettime(timing_us0)
#endif

      !! create new partitioning, myblocks contains my local partition
      !! creates new domain%blkmap
      
      
      !! ParMETIS
      
#ifdef FD4_METIS
      if(meth==FD4_BALANCE_METIS .and. allblocks%pos>0) then
        ubvec(:) = 1.15_r4k !TODO: tune it, or try to guess a good value dynamically
        itr = 10.0_r4k      !TODO: tune it, or try to guess a good value dynamically
        subset = -1         ! means: no subset
        wfactor = 1.0
        if(present(opt_ubvec))   ubvec(:) = opt_ubvec
        if(present(opt_itr))     itr = opt_itr
        if(present(opt_subset))  subset = opt_subset
        if(present(opt_wfactor)) wfactor = opt_wfactor
        if(subset<2) then
          call fd4_part_metis(domain, allblocks%a, allblocks%pos, myblocks, newpartsize, ubvec, itr, wfactor)
        else
          call fd4_part_metis_subset(domain, allblocks%a, allblocks%pos, myblocks, newpartsize, ubvec, itr, subset)
        end if
      end if
#endif
      

      !! Hilbert SFC

      if(meth==FD4_BALANCE_HSFC_BINSRCH  .or. meth==FD4_BALANCE_HSFC_SIMPLE .or. &
         meth==FD4_BALANCE_HSFC_PARALLEL .or. allblocks%pos==0) then
        if(use_weights) then
          call fd4_part_sfc_weight(domain, allblocks%a, allblocks%pos, myblocks, newpartsize, meth, err, blocks_changed)
          FD4_L1_ASSERT(err/=0,'fd4_part_sfc_weight failed')
        else
          call fd4_part_new(domain, allblocks%a, allblocks%pos, myblocks, newpartsize, blocks_changed)
        end if
      end if


      !! serial (for debugging)
      
      if(meth==FD4_BALANCE_SERIAL .and. allblocks%pos>0) then
        !! a dummy method which moves all blocks to rank 0
        ! call sfc partitioning to update the blkmap
        call fd4_part_new(domain, allblocks%a, allblocks%pos, myblocks, newpartsize)
        ! now move all blocks to rank 0
        if(mpi%rank==0) then
          myblocks = allblocks%a
          newpartsize = allblocks%pos
        else
          newpartsize = 0
        end if      
        ! and set all blocks in blkmap to 0
        where( domain%blkmap( :, :, :, domain%blkmap_now ) > 0 ) domain%blkmap( :, :, :, domain%blkmap_now ) = 0
      end if


#ifdef FD4_PARTITION_STATISTICS  
      call gettime(timing_us1)
      balance_timing_us = timing_us1 - timing_us0
      timing_us0 = timing_us1
#endif


      if(present(opt_stats)) then
        opt_stats%partition_changed = .true.
        opt_stats%migrated_blocks = COUNT (  &
            domain%blkmap( 1:domain%blnum(1), 1:domain%blnum(2), 1:domain%blnum(3), domain%blkmap_now  ) /=  &
            domain%blkmap( 1:domain%blnum(1), 1:domain%blnum(2), 1:domain%blnum(3), domain%blkmap_last )  )
        if(blocks_changed) then
          opt_stats%created_blocks = COUNT (  &
              domain%blkmap( 1:domain%blnum(1), 1:domain%blnum(2), 1:domain%blnum(3), domain%blkmap_now  ) >= 0 .and. &
              domain%blkmap( 1:domain%blnum(1), 1:domain%blnum(2), 1:domain%blnum(3), domain%blkmap_last ) <  0 )
          opt_stats%removed_blocks = COUNT (  &
              domain%blkmap( 1:domain%blnum(1), 1:domain%blnum(2), 1:domain%blnum(3), domain%blkmap_now  ) <  0 .and. &
              domain%blkmap( 1:domain%blnum(1), 1:domain%blnum(2), 1:domain%blnum(3), domain%blkmap_last ) >= 0 )
          opt_stats%migrated_blocks = opt_stats%migrated_blocks - opt_stats%removed_blocks - opt_stats%created_blocks
        end if
      end if


!!
!! 3. block migration
!!

      call fd4_balance_block_migration(domain, mystack, myblocks, newpartsize, err, opt_stats)
      FD4_L3_ASSERT(err/=0,'fd4_balance_block_migration failed')


      deallocate(myblocks)


      domain%blockcount = allblocks%pos

#ifdef FD4_PARTITION_STATISTICS  
    else
      call gettime(timing_us0)
#endif

    end if  !! needed new partitioning

    
    call stack_delete(mystack)
    

!!
!! 4. communication ghost block update
!!

    ! only run ghost block update if partitioning changed or forced by user
    if(part_changed .or. force) then

      call fd4_balance_update_ghosts(domain, opt_stats)
      
    end if

#ifdef FD4_PARTITION_STATISTICS  
      call gettime(timing_us1)
      migration_timing_us = timing_us1 - timing_us0
#endif

    FD4_L5_INFO('number of comm. ghostblocks: '//i2s(domain%glist(FD4_GHOST_CMM)%length))

    ! write status message
    if(present(opt_stats)) then
      if(part_changed) then
        if(blocks_changed) then
          if(opt_stats%migrated_blocks>0) then
            opt_stats%status = 'blocks changed, migrated blocks'
          else
            opt_stats%status = 'blocks changed, no migration'
          end if
        else
          if(opt_stats%migrated_blocks>0) then
            opt_stats%status = 'not balanced, migrated blocks'
          else
            opt_stats%status = 'not balanced, no migration'
          end if
        end if
      else
        if(.not. lb_benefit) then
          opt_stats%status = 'no benefit (auto mode)'
        else
          opt_stats%status = 'sufficient balance'
        end if        
      end if
    end if

    ! memorize if partitioning has been (tried to) change
    domain%balance%part_changed = part_changed

    if(part_changed .and. auto) then

      ! clock the end of this routine
      call gettime(lbtime1)

      ! only put timings to next record, if we already have globalized the current record
      ! (otherwise the timings would not be globalized and we could get a deadlock!)
      if(.not. domain%balance%globalize_lbtime) then
        domain%balance%curlbtime = MOD(domain%balance%curlbtime, FD4_LB_HISTORY ) + 1
      end if      
      ! put the time into the array of last timings in domain%balance
      domain%balance%lbtimes(domain%balance%curlbtime) = MAX( REAL( lbtime1 - lbtime0, r4k ) , 1.0_r4k )
      ! indicate that this time needs to be globalized (max of all ranks) in next fd4_balance_readjust
      domain%balance%globalize_lbtime = .true.

    end if


#ifdef FD4_BALANCE_DUMP_WEIGHTS
    ! dump weights after timing, so that dumping does not influence the measured time
    if(part_changed) then
      if(mpi%rank==0) then
         dumpcounter=dumpcounter+1
         write(1000+dumpcounter,'(3(I7))') domain%blnum(1:3)
         write(1000+dumpcounter,'(6(I7))') domain%dbnd(1:3,1:2)
         do k=1,domain%blnum(3)
           do j=1,domain%blnum(2)
             do i=1,domain%blnum(1)
               write(1000+dumpcounter,'(E15.8)') domain%balance%blkweights( i, j, k )
             end do
           end do
         end do
      end if
      call MPI_Barrier(mpi%comm, err)
    end if
#endif
    
    err = 0
    
  end subroutine fd4_balance_readjust




  !! Update of global empty info ($domain%$empties)
  !!  - update empty info of local blocks ($sempty)
  !!  - send number of changes in $sempty to all other processes (MPI_Allgather)
  !!  - if there's no change on any process
  !!    - return with $status=.false.
  !!  - exchange of the changes in $sempty as pair ($id, $sempty) using MPI_Allgatherv
  !!  - update $domain%$empties
  !!  - return with $status=.true.
  !!
  !! The $myempties stack will be returned allocated and can be re-used after the call.
  !! If not, $myempties should be deallocated right after this function.
  !!
  !! Internal routine for $fd4_balance_readjust.
  subroutine fd4_balance_global_empty_update(domain, myempties, status, starttime)
    type(fd4_domain), intent(inout), target :: domain     !! the domain
    type(stack),      intent(inout)         :: myempties  !! stack for internal use
    logical,          intent(out)           :: status     !! return status (true if empty info changed)
    integer(i8k),     intent(out)           :: starttime  !! time after first synchronizing global communication
    
    type(fd4_iter) :: iter
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: idx, st, i, j, k, nb
    integer(i4k) :: e
    integer(i_k), allocatable :: allempties(:)     ! for MPI_Allgatherv'ing all block%sempty changes
    integer :: err
    
    status = .false.
    mpi => domain%mpi


    call stack_create(myempties,domain%blocks%length)
    
    
    !! summarize empty-info of each block
    
    ! loop over blocks
    iter%cur => rbtree_fd4_block_iterator_first(iter%biter, domain%blocks)
    do while(associated(iter%cur))
            
      e = 0_i4k
      ! summarize empty-info of all variables to e
      do idx=1,size(domain%vartab)
        !TODO: save us from the if by creating an array of vars with threshold and include other necessary information
        if(domain%vartab(idx)%vthres == FD4_NOTHRES) cycle
        do st=1,domain%vartab(idx)%nsteps
          e = IOR(iter%cur%empty(idx,st), e)
        end do
      end do
      
      if(e /= iter%cur%sempty) then
        iter%cur%sempty = e
        domain%empties(iter%cur%pos(1),iter%cur%pos(2),iter%cur%pos(3)) = e
        !                  block position (index) ---,  ,--- empty state
        call stack_push2(myempties, iter%biter%cur%key, e)
      end if

      iter%cur => rbtree_fd4_block_iterator_next(iter%biter)
    end do
    
    
    !! tell all other processes how many changes I see in my local blocks
    call MPI_Allgather( myempties%pos, 1, MPI_I_K, mpi%recvcnt, 1, MPI_I_K, mpi%comm, err )
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allgather failed')
    
    call gettime(starttime)    
    
    !! let's see if anything has changed, if no, we can leave this routine
    nb = SUM(mpi%recvcnt)
    if(nb == 0) then
      return
    end if
    
    
    !! now collect the changed empty information from those processes who claimed they have changed blocks
    
    ! set receive displacements for MPI_Allgatherv
    mpi%recvdis(0) = 0
    do i=0,mpi%nprocs-2
      mpi%recvdis(i+1) = mpi%recvdis(i) + mpi%recvcnt(i)
    end do

    allocate(allempties(nb))

    ! gather the empty info from all processes
    call MPI_Allgatherv (myempties%a, myempties%pos,              MPI_I_K, &
                         allempties,  mpi%recvcnt,   mpi%recvdis, MPI_I_K, mpi%comm, err )
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')
    
    ! this is the number of blocks with changed empty info
    nb = nb/2   

    !! update empty info of changed blocks in domain%empties
    do idx=1,nb
      ! calculate position from id
      call fd4_domain_i2ijk(allempties(2*idx-1), i, j, k)
      ! update empty info
      domain%empties(i,j,k) = INT(allempties(2*idx),i4k)
      ! apply periodic boundaries to domain%empties
      if(domain%periodic(1)) then
        if(i==1)               domain%empties(domain%blnum(1)+1,j,k) = domain%empties(i,j,k)
        if(i==domain%blnum(1)) domain%empties(0                ,j,k) = domain%empties(i,j,k)
      end if
      if(domain%periodic(2)) then
        if(j==1)               domain%empties(i,domain%blnum(2)+1,k) = domain%empties(i,j,k)
        if(j==domain%blnum(2)) domain%empties(i,0                ,k) = domain%empties(i,j,k)
      end if
      if(domain%periodic(3)) then
        if(k==1)               domain%empties(i,j,domain%blnum(3)+1) = domain%empties(i,j,k)
        if(k==domain%blnum(3)) domain%empties(i,j,0                ) = domain%empties(i,j,k)
      end if
    end do
    
    deallocate(allempties)

    status = .true.

  end subroutine fd4_balance_global_empty_update


  !! Block migration.
  !!
  !!  - allocate new blocks and start receiving
  !!  - send blocks
  !!  - deallocate blocks (not needed anymore or migrated)
  !!
  !! The $rmblock stack will be returned allocated and can be re-used after the call.
  !! If not, $rmblock should be deallocated right after this function.
  !!
  !! Internal routine for $fd4_balance_readjust.
  subroutine fd4_balance_block_migration(domain, rmblock, myblocks, newpartsize, err, opt_stats)
    type(fd4_domain), intent(inout), target :: domain        !! the domain
    type(stack),      intent(inout)         :: rmblock       !! stack for internal use
    integer(i_k),     intent(in)            :: myblocks(:,:) !! coordinates of local blocks (result from partitioning algorithm)
    integer(i_k),     intent(in)            :: newpartsize   !! size of myblocks(3,:)    
    integer(i_k),     intent(out)           :: err           !! error status: 0...ok
    type(fd4_balance_statistics), &
                      intent(inout),optional:: opt_stats     !! statistics about migration

    type(fd4_iter) :: iter
    type(fd4_block), pointer :: b
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: pos(3), bidx, i
    integer :: crank, request

    type fd4_i4k_link
      integer(i4k), pointer :: l => null()
    end type fd4_i4k_link
    type(fd4_i4k_link), allocatable :: mindispl(:)
    type(fd4_mpitype), allocatable ::  migrationtype(:)
    integer(MPI_ADDRESS_KIND) :: disp
    integer :: bl_type, toff
    logical :: is_min
    integer(i_k) :: block_allocations, block_deallocations
    
    block_allocations = 0
    block_deallocations = 0
    
    err = -1
    mpi => domain%mpi

    !! start sending data of migrated blocks in the new partitioning

    ! allocate stack to useful size
    call stack_create(rmblock,newpartsize)

    ! reset the stack of requests
    mpi%requests%pos = 0        

    toff = 0
#ifdef FD4_WAIT_BEFORE_TYPE_FREE
    toff = mpi%nprocs
#endif
    allocate(mindispl(0:mpi%nprocs-1))
    allocate(migrationtype(0:mpi%nprocs-1+toff))

#ifdef FD4_PARTITION_STATISTICS
    comm_total_size = -1
#endif

    ! loop over blocks
    iter%cur => rbtree_fd4_block_iterator_first(iter%biter, domain%blocks)
    do while(associated(iter%cur))
      crank = domain%blkmap( iter%cur%pos(1), iter%cur%pos(2), iter%cur%pos(3), domain%blkmap_now )
      ! if   crank == -1:       block not present in new partitioning -> block can be deleted
      ! elif crank == mpi%rank: block is present in last and current partitioning ->  nothing to do
      ! else :                  migrated block -> need to send data to crank and delete block later on
      if(crank /= mpi%rank) then
        ! don't delete the block during this iteration:
        ! * this causes trouble with the iterator
        ! * we need to keep the block until the send has finished (waitall below)
        call stack_push(rmblock, iter%biter%cur%key)
        if(crank /= -1) then
          ! start sending the block
          b => rbtree_fd4_block_get(domain%blocks, iter%biter%cur%key)
          call fd4_comm_build_type(b, bl_type, err, addr = disp)
          FD4_L3_ASSERT(err/=0,'fd4_comm_build_type failed')
          call fd4_mpitype_add( migrationtype(crank), INT(iter%biter%cur%key,8), 1, disp, bl_type, is_min)
          if(is_min) mindispl(crank)%l => b%empty(1,1)
          FD4_L5_INFO('sending block '//i2s(iter%biter%cur%key)//' to rank '//i2s(crank))
#if ( FD4_VERBOSE_LEVEL >= 4 )
        else if(domain%partition_id>1_i8k) then
          FD4_L4_INFO('deleting block '//i2s(iter%biter%cur%key))
#endif
        end if
      end if
      iter%cur => rbtree_fd4_block_iterator_next(iter%biter)
    end do


    do i=0,mpi%nprocs-1
      if(migrationtype(i)%count>0) then
        call fd4_mpitype_commit(migrationtype(i), err)
        FD4_L3_ASSERT(err/=0,'fd4_mpi_datatype_commit failed at '//i2s(i))
        call MPI_Isend(mindispl(i)%l, 1, migrationtype(i), i, migrationtype(i)%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
        call stack_push(mpi%requests, request)
#ifndef FD4_WAIT_BEFORE_TYPE_FREE
        call fd4_mpitype_free(migrationtype(i), .true., .true.)
#endif
      end if
    end do


#ifdef FD4_PARTITION_STATISTICS
    comm_total_size = 0
#endif

    !! allocate new blocks & start receiving data

    do i=1,newpartsize
      pos(1:3)=myblocks(1:3,i)
      call fd4_domain_c2i(pos, bidx)
      if( .not. associated(rbtree_fd4_block_get(domain%blocks,bidx)) ) then
        crank = domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_last)
        call fd4_domain_add_block(domain, pos, crank<0, err, b, opt_alloc = block_allocations, opt_dealloc = block_deallocations)
        FD4_L3_ASSERT(err/=0,'fd4_domain_add_block failed')
        ! if crank == -1: new created block
        !    crank >=  0: migrated block, need to receive data from crank
        FD4_L3_ASSERT(crank==mpi%rank,'inconsistency in blkmap '//i2s(bidx))
        if(crank >= 0) then
          ! add this block to crank's type
          call fd4_comm_build_type(b, bl_type, err, addr = disp)
          FD4_L3_ASSERT(err/=0,'fd4_comm_build_type failed')
          call fd4_mpitype_add( migrationtype(crank+toff), INT(bidx,8), 1, disp, bl_type, is_min)
          if(is_min) mindispl(crank)%l => b%empty(1,1)
          FD4_L5_INFO('receiving block '//i2s(bidx)//' from rank '//i2s(crank))
        end if
      end if          
    end do


    do i=0,mpi%nprocs-1
      if(migrationtype(i+toff)%count>0) then
        call fd4_mpitype_commit(migrationtype(i+toff), err)
        FD4_L3_ASSERT(err/=0,'fd4_mpi_datatype_commit failed at '//i2s(i))
        call MPI_Irecv(mindispl(i)%l, 1, migrationtype(i+toff), i, migrationtype(i+toff)%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
        call stack_push(mpi%requests, request)
#ifndef FD4_WAIT_BEFORE_TYPE_FREE
        call fd4_mpitype_free(migrationtype(i+toff), .true., .true.)
#endif
      end if
    end do

    !! wait for the transfer to complete

    if(mpi%requests%pos>0) then
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
    end if    

#ifdef FD4_WAIT_BEFORE_TYPE_FREE
    do i=0,mpi%nprocs-1
      if(migrationtype(i)%count>0)      call fd4_mpitype_free(migrationtype(i)     , .true., .true.)
      if(migrationtype(i+toff)%count>0) call fd4_mpitype_free(migrationtype(i+toff), .true., .true.)
    end do
#endif

    deallocate(migrationtype)
    deallocate(mindispl)    

    !! deallocate old blocks
    do i=1,rmblock%pos
      call fd4_domain_delete_block(domain, rmblock%a(i), err, opt_dealloc = block_deallocations)
      FD4_L3_ASSERT(err/=0,'delete block failed')
    end do
    
    if(present(opt_stats)) then
      opt_stats%allocated_blocks   = opt_stats%allocated_blocks   + block_allocations
      opt_stats%deallocated_blocks = opt_stats%deallocated_blocks + block_deallocations
    end if
    
    err = 0
    
  end subroutine fd4_balance_block_migration




  !! Communication / boundary ghost block update
  !!
  !!  - Allocate comm. ghost blocks and put them into the ghostlist
  !!  - Allocate the ghost if off-processer neighbor block is present
  !!    without caring for threshold
  !!  - This makes results independent from the partitioning
  !!  - Additionally, update the spatial union of bounds of all local blocks ($domain%$blockunion)
  !!
  !! Notes:
  !!  - (1) Ghost blocks may be added or removed.
  !!  - (2) And even if there has been no ghost block added or removed, the
  !!        neighbor may have changed which is important to notice for ghost
  !!        communiation.
  !! These issues are tracked with the domain%glist(FD4_GHOST_CMM)%touched flag (1) and
  !! the local variable p_changed (2). If one of them is true after this
  !! loop, we call the (local!) partitioning "new" (by counting up the
  !! local partition_id). In this case, the ghost communication has to be
  !! adapted to the new situation.
  !!
  !! Internal routine for $fd4_balance_readjust.
  subroutine fd4_balance_update_ghosts(domain, opt_stats)
    type(fd4_domain), intent(inout), target :: domain        !! the domain
    type(fd4_balance_statistics), &
                      intent(inout),optional:: opt_stats     !! statistics about ghost (de)allocation
                      
    type(fd4_iter) :: iter
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: dim, dir, pos(3)
    integer :: crank
    logical :: p_changed
    integer(i_k) :: block_allocations, block_deallocations
    
    block_allocations = 0
    block_deallocations = 0
    
    mpi => domain%mpi
    p_changed = .false.
    
    ! reset spatial union of bounds of all local blocks
    call section_reset(domain%blockunion, domain%dbnd)

    ! loop over blocks
    iter%cur => rbtree_fd4_block_iterator_first(iter%biter, domain%blocks)
    do while(associated(iter%cur))

      ! add block's global grid bounds to the union of all local blocks
      call section_add(domain%blockunion, iter%cur%bnd)

      do dim=1,3
        do dir=1,2
        
          ! neighbor block's position
          pos(:)   = iter%cur%pos
          pos(dim) = pos(dim) + fd4_incdec(dir)
          
          crank = domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now)
          
          if(crank<0) then
            ! nobody has allocated the neighbor
            ! so let's remove the comm. ghost block if there is any (from last partitioning)
            if(associated(iter%cur%neigh(dim,dir)%l)) then
              if( iter%cur%neigh(dim,dir)%l%pos(1) == FD4_GHOST_CMM ) then
                FD4_L5_INFO('removing ghost block '//p2s(iter%cur%pos)//' - '//p2s(pos))
                call fd4_domain_delete_ghost(domain,iter%cur%neigh(dim,dir)%l, opt_dealloc = block_deallocations)
                iter%cur%neigh(dim,dir)%l => NULL()
              end if
            end if
#ifndef FD4_INSTANT_BGB
            ! check boundary ghost blocks
            if(.not. domain%periodic(dim) .and. (pos(dim)<1 .or. pos(dim)>domain%blnum(dim))) then
              !if( IAND( IOR(iter%cur%nempty,iter%cur%sempty) , BLOCK_EMPTY(dim,dir) ) > 0 ) then
              if( IAND( iter%cur%sempty , BLOCK_EMPTY(dim,dir) ) > 0 ) then
                ! we need a bgb
                if(.not. associated(iter%cur%neigh(dim,dir)%l)) then
                  ! add bgb
                  call fd4_domain_add_ghost(domain, iter%cur, dir, dim, FD4_GHOST_BND, opt_alloc = block_allocations)
                end if
              else
                ! we don't need a bgb
                if(associated(iter%cur%neigh(dim,dir)%l)) then
                  ! remove bgb
                  call fd4_domain_delete_ghost(domain,iter%cur%neigh(dim,dir)%l, opt_dealloc = block_deallocations)
                  iter%cur%neigh(dim,dir)%l => NULL()
                end if
              end if
            end if
#endif
          elseif(crank>=0 .and. crank/=mpi%rank) then
            ! another rank has allocated the neighbor...
            ! do we already have a comm. ghost here?
            if(.not. associated(iter%cur%neigh(dim,dir)%l)) then
              ! so let's add a comm. ghost block
              FD4_L5_INFO('adding a ghost block '//p2s(iter%cur%pos)//' - '//p2s(pos))
              call fd4_domain_add_ghost(domain, iter%cur, dir, dim, FD4_GHOST_CMM, opt_alloc = block_allocations)
            elseif( crank /= domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_last) ) then
              ! we already have a ghost here, but the neighbor has changed.
              ! we need this information for ghost communication
              FD4_L5_INFO('new neighbor '//p2s(iter%cur%pos)//' - '//p2s(pos))
              p_changed = .true.
            end if
          end if
          
        end do ! dir
      end do ! dim

      iter%cur => rbtree_fd4_block_iterator_next(iter%biter)
    end do
    

    ! count up the id of this (local) partition if the partition has changed
    if(domain%glist(FD4_GHOST_CMM)%touched .or. p_changed) then
      domain%partition_id = domain%partition_id + 1
      FD4_L4_INFO('new partition, id = '//l2s(domain%partition_id))
      ! reset ghost list's touched flag
      domain%glist(FD4_GHOST_CMM)%touched = .false.    
    end if
    
    if(present(opt_stats)) then
      opt_stats%allocated_blocks   = opt_stats%allocated_blocks   + block_allocations
      opt_stats%deallocated_blocks = opt_stats%deallocated_blocks + block_deallocations
    end if
    
  end subroutine fd4_balance_update_ghosts


  !! Determine the estimated load balance using block weigths.
  !!
  !! TODO: use only a single MPI_Allreduce
  !!
  !! Internal routine for $fd4_balance_readjust.
  subroutine fd4_balance_estimate(domain, auto, lb_benefit, balance, starttime, opt_stats)
    type(fd4_domain),      intent(inout) :: domain     !! the domain
    logical,               intent(in)    :: auto       !! automatic decision whether load balancing beneficial or not
    logical,               intent(out)   :: lb_benefit !! if auto==.true.: is load balancing beneficial?
    real(r4k),             intent(out)   :: balance    !! the load balance estimate: 1 = perfect, 0 = infinitely poor
    integer(i8k),optional, intent(out)   :: starttime  !! time after first synchronizing global communication
    type(fd4_balance_statistics), &
                 optional, intent(out)   :: opt_stats  !! return some statistics about load balancing

    type(fd4_iter) :: iter
    real(r4k) :: myload     ! load of local process
    real(r4k) :: maxload    ! load of the the processes with max load
    real(r4k) :: sumload    ! total load of the processes
    real(r4k) :: myin(2), maxout(2) ! for combining allreduce for load an time
    integer :: err, nmax
    real(r4k) :: tbalance, tlost, bestbalance
    
    ! default
    lb_benefit = .true.
    
    balance = 0.0
    myload = 0.0
    
    ! loop over blocks
    iter%cur => rbtree_fd4_block_iterator_first(iter%biter, domain%blocks)
    do while(associated(iter%cur))
      myload = myload + iter%cur%weight
      iter%cur => rbtree_fd4_block_iterator_next(iter%biter)
    end do
    
    ! determine max process load and the sum of all total loads 
    !TODO: do this with a single MPI_Allreduce (projekt/fd4/test/mpi_reduce_op)
    !TODO: where put the mpi op routine? -> in this module
    !TODO: where intialize the routine? -> in 1. call, use module variable to indicate routine is initialized
    call MPI_Allreduce(myload, sumload, 1, MPI_R4K, MPI_SUM, domain%mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allreduce failed')

    if(present(starttime)) call gettime(starttime)
    
    nmax = 1
    myin(1) = myload
    
    if( auto .and. domain%balance%globalize_lbtime ) then
      ! we also want to compute the max of the current lbtime of all processes
      ! (this is the load balancing timing from the last call)
      myin(2) = domain%balance%lbtimes(domain%balance%curlbtime)
      nmax = 2
    end if

    call MPI_Allreduce(myin, maxout, nmax, MPI_R4K, MPI_MAX, domain%mpi%comm, err)
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allreduce failed')
    
    maxload = maxout(1)
    
    if(maxload>0) then
      balance = sumload / ( maxload * domain%mpi%nprocs )
    else
      balance = 1.0_r4k
    endif
    
    if(domain%balance%part_changed) then
      domain%balance%curlbbalance = MOD(domain%balance%curlbbalance, FD4_LB_HISTORY ) + 1
      domain%balance%lbbalance(domain%balance%curlbbalance) = balance
      domain%balance%part_changed = .false.
    end if
    
    
    if(nmax > 1) then
      ! set the current lbtime to the max of all times over all processes
      domain%balance%lbtimes(domain%balance%curlbtime) = maxout(2)
      domain%balance%globalize_lbtime = .false.
    end if
    
    ! automatic mode: check if it is worth spending the time for load balancing
    ! (lost time due to imbalance vs. runtime of load balancing)
    if( auto ) then
      ! we only do this, if half of the lb times history is filled, but at least 3
      if( count(domain%balance%lbtimes > 0.0) >= MAX( FD4_LB_HISTORY/2, 3 ) ) then
        ! estimated time for load balancing
        tbalance = MINVAL( domain%balance%lbtimes , MASK = domain%balance%lbtimes > 0.0 )
        ! best balance achieved after last calls to fd4_balance_readjust
        bestbalance = MAXVAL( domain%balance%lbbalance , MASK = domain%balance%lbbalance > 0.0 )
        ! estimated time lost due to imbalance
        !       ,-- run time      ,-- run time if best balance (optimistic)
        tlost = maxload - sumload / ( domain%mpi%nprocs * MAX( bestbalance, balance ) )
        lb_benefit = ( tlost > tbalance )
        if(FD4_VERBOSE_LEVEL >= 4 .and. domain%mpi%rank == 0) then
          write (fd4_out,'(A,A,F12.2,A,F12.2)') FD4_LINEHD,' tlost = ',tlost,'  tbalance = ',tbalance
        end if
        if(present(opt_stats)) then
          opt_stats%auto_tlost = tlost
          opt_stats%auto_tbalance = tbalance
        end if
      end if
    end if
    
    
    
  end subroutine fd4_balance_estimate
  
  
  
  !! Find out which blocks are required.
  !!
  !! Creates domain%balance%allblocks, array of blocks (positions) which are needed in the
  !! new partitioning using the empties array.
  !! This might be costly when using a large number of blocks.
  subroutine fd4_balance_required_blocks(domain, blocks_changed)
    type(fd4_domain), intent(inout), target :: domain          !! the domain
    logical,          intent(out)           :: blocks_changed  !! any blocks created or removed?
    
    type(stack3), pointer :: allblocks             ! coordinates of all required blocks in global domain
    integer(i4k) :: e
    integer(i_k) :: i, j, k, pos(3), idx
    
    blocks_changed = .false.
    allblocks => domain%balance%allblocks
  
    ! first, look if there are marked blocks
    if( domain%blocks_are_marked ) then
      call stack3_create(allblocks,domain%markedblocks%pos+10)
      do i=1,domain%markedblocks%pos
        call fd4_domain_i2c(domain%markedblocks%a(i),allblocks%a(:,i))
        blocks_changed = blocks_changed .or. &
           (domain%blkmap(allblocks%a(1,i),allblocks%a(2,i),allblocks%a(3,i),domain%blkmap_now) < 0)
      end do
      allblocks%pos = domain%markedblocks%pos
    else
      call stack3_create(allblocks,domain%blocks%length+10)
    end if

    ! loop over domain%empties(:,:,:)
    do k=1,domain%blnum(3)
      pos(3) = k
      do j=1,domain%blnum(2)
        pos(2) = j
        i_loop: do i=1,domain%blnum(1)
          pos(1) = i
          ! check if block is not empty
          if(domain%empties(i,j,k)>0) then
            if(domain%blocks_are_marked) then
              call stack_binary_search(domain%markedblocks, fd4_domain_c2if(pos), domain%markedblocks%pos, idx)
              if(idx/=0) cycle i_loop 
            end if
            call stack3_push(allblocks, pos)              
            blocks_changed = blocks_changed .or. (domain%blkmap(i,j,k,domain%blkmap_now) < 0)
            cycle i_loop
          end if          
          ! check neighbors, maybe one of them requires this block
          e =        IAND( domain%empties(i  ,j  ,k-1), BLOCK_EMPTY(3,2) )
          e = IOR(e, IAND( domain%empties(i  ,j-1,k  ), BLOCK_EMPTY(2,2) ) )
          e = IOR(e, IAND( domain%empties(i-1,j  ,k  ), BLOCK_EMPTY(1,2) ) )
          e = IOR(e, IAND( domain%empties(i+1,j  ,k  ), BLOCK_EMPTY(1,1) ) )
          e = IOR(e, IAND( domain%empties(i  ,j+1,k  ), BLOCK_EMPTY(2,1) ) ) 
          e = IOR(e, IAND( domain%empties(i  ,j  ,k+1), BLOCK_EMPTY(3,1) ) ) 
          if( e /= 0 ) then
            if(domain%blocks_are_marked) then
              call stack_binary_search(domain%markedblocks, fd4_domain_c2if(pos), domain%markedblocks%pos, idx)
              if(idx/=0) cycle i_loop 
            end if
            call stack3_push(allblocks, pos)              
            blocks_changed = blocks_changed .or. (domain%blkmap(i,j,k,domain%blkmap_now) < 0)
            cycle i_loop
          end if
        end do i_loop
      end do
    end do

    blocks_changed = blocks_changed .or. ( domain%blockcount /= allblocks%pos )

    if( domain%blocks_are_marked ) then
      domain%markedblocks%pos = 0  ! reset the stack, but keep memory
      domain%blocks_are_marked = .false.
    end if

  end subroutine fd4_balance_required_blocks


end module fd4_balance_mod
